/*
**** ingproc.sc
**  = An Ingres backend for GSQL
**  
**  The functions contained in this file are intended as 
**  substitutes for those in the 'sybproc.c' file available
**  through Jason Ng at ncsa.
**
**  ingproc.sc has been tested under Solaris 2.3
**
**  Produced by:	Gordon Gallacher
**			King's College London
**
**			g.gallacher@bay.cc.kcl.ac.uk
**
**			November, 1994
**
**  (Having said that, Ingres people will easily recognise the
**   bulk of the code as having been lifted from the terminal 
**   monitor example in the ESQL/C manual.  However, I've 
**   simplified that slightly to use a select loop rather than a 
**   cursor.  Also, there is a potential bug in the code in that
**   manual: nullable columns were not were not being aligned
**   in the result buffer area, leading (at least on Solaris) to
**   a core dump.  I've fixed that. )
**
**  NB: Note that I make use of 'user defines' in the .proc file
** 	to pass certain information.  Specifically, you must
**  	define your local value of II_SYSTEM, and the name of
**	database the .proc file refers to.  For example:
**
**	DEFINE II_SYSTEM II_SYSTEM=/usr/local
**	DEFINE DBNAME whatever
**
**  Another user-define I have implemented here is a method for
**  making columns invisible.  Ie, if you need to include a column
**  in the select list, because of an 'order by', but don't really
**  want that column to be displayed to the user, then use this
**  define in the .proc file:
**
**    DEFINE <column-name>_invisible TRUE;
**
**  (It doesn't actually matter what definition you give; if anything 
**   at all is there, it'll work.)
**
**  The 'make.ingback' script should be used for creating 'sqlmain'.
**
*/

#include <stdio.h>
#include <sys/ddi.h>

EXEC SQL INCLUDE SQLCA;
EXEC SQL INCLUDE SQLDA;

EXEC SQL DECLARE stmt STATEMENT;

#define	DEF_ELEMS	8
#define	DATE_SIZE	25

#define SQL_NOTFOUND	100

#define	DBNAME_MAX	50
#define	INPUT_SIZE	256
#define	STMT_MAX	1000

IISQLDA	*sqlda = (IISQLDA *) 0;   /* point to SQL dynamic area	*/

/* define storage for results  */

struct {
	int res_length;
	char	*res_data;
} res_buf = {0, NULL};

int	colWidth[50];	/* holds column widths for printing	*/
char	nullMessage[] = "n/a ";
char	defName[40];
char	*defValue;
int	invisible[100];

void Init_Sqlda();
void Print_Header();
void Init_Ingres();
void Print_Row();
char *Alloc_Mem();
char *calloc();
void Print_Error();


void Init_Sqlda(num_elems)
int num_elems;
{

  /* if we already have one, free it 	*/

  if (sqlda)
	free( (char *) sqlda);

  sqlda = (IISQLDA *) Alloc_Mem(IISQDA_HEAD_SIZE + (num_elems *
					IISQDA_VAR_SIZE), "new SQLDA");
  sqlda->sqln = num_elems;

}


void Init_Ingres(char *dbname)
{

  EXEC SQL BEGIN DECLARE SECTION;

	char v_dbname[DBNAME_MAX+1];

  EXEC SQL END DECLARE SECTION;
 
  putenv( (char *) app_getdefine("II_SYSTEM") );
  /* printf(" Env: %s\n", getenv("II_SYSTEM"));  */
  EXEC SQL WHENEVER SQLERROR call Print_Error;
  strcpy (v_dbname, dbname); 
  EXEC SQL CONNECT :v_dbname;

}


int execute_sql(char *sqlstring)
{

  EXEC SQL BEGIN DECLARE SECTION;
	char v_stmt_buf[500];
  EXEC SQL END DECLARE SECTION;

  int rows;

#define MARGIN 2
#define STRSIZE 80
#define LONGSTRSIZE 300

  Init_Ingres( (char *)app_getdefine("DBNAME"));

 /*  printf("<PRE>SQL string: %s\n</PRE>", sqlstring); */

  strcpy(v_stmt_buf, sqlstring);
  /* allocate an sqlda ...	*/

  Init_Sqlda(DEF_ELEMS);	/* start with DEF_ELEMS cols.	*/
				/* (we'll expand if necessary)	*/

  EXEC SQL WHENEVER SQLERROR GOTO Stmt_err;

  /* prepare the SQL statement 	*/

  EXEC SQL PREPARE stmt FROM :v_stmt_buf;
  EXEC SQL DESCRIBE stmt INTO :sqlda;

  if (sqlda->sqld > sqlda->sqln)
  {
	/* printf("sqld, sqln: %d %d \n", sqlda->sqld, sqlda->sqln); */
	Init_Sqlda(sqlda->sqld);
	EXEC SQL DESCRIBE stmt INTO :sqlda;
  }

  Print_Header();
  rows = 0;
  printf("<PRE>"); 

  /* ok, we're ready to execute the select;  this is now a select loop,
  ** replacing the cursor loop in the tm example.
  */

  EXEC SQL EXECUTE IMMEDIATE :v_stmt_buf
	USING DESCRIPTOR :sqlda;
  EXEC SQL BEGIN;

	rows++;  
	Print_Row();  
	printf("\n");

  EXEC SQL END;

  printf("</PRE>"); 
  EXEC SQL disconnect;

  return 0;

 Stmt_err:
  EXEC SQL WHENEVER SQLERROR continue;
  Print_Error();
  EXEC SQL disconnect;
  return 1;
}


void Print_Header()
{

  /* not only print the headers, but also figure some sizes	*/
  /* for the columns, and initialise sqlda pointers to buffer	*/

  int i;		/* pointer to sqlvar	*/
  IISQLVAR	*sqv;
  int base_type;
  int res_cur_size;
  int round;

  /* for each column, print the title, and figure	*/
  /* total buf size needed for result	*/

  printf("<PRE>"); 
  for (res_cur_size = 0, i = 0; i < sqlda->sqld; i++)
  {

	sqv = &sqlda->sqlvar[i];

	/* get the base type 	*/

 	if ((base_type = sqv->sqltype) < 0)
		base_type = -base_type;

	/* collapse types to Int, Float, or Char	*/

	switch (base_type)
	{
	  case IISQ_INT_TYPE:
		res_cur_size += sizeof(long);
		sqv->sqllen = sizeof(long);
		break;

	  case IISQ_MNY_TYPE:
		if (sqv->sqltype <0)
		  sqv->sqltype = -IISQ_FLT_TYPE;
		else
		  sqv->sqltype = IISQ_FLT_TYPE;

		res_cur_size += sizeof(double);
		sqv->sqllen = sizeof(double);
		break;

	  case IISQ_FLT_TYPE:
		res_cur_size += sizeof(double);
		sqv->sqllen = sizeof(double);
		break;

	  case IISQ_DTE_TYPE:
		sqv->sqllen = DATE_SIZE;

	  case IISQ_CHA_TYPE:
	  case IISQ_VCH_TYPE:
		res_cur_size += sqv->sqllen + 1;
		/* round off...*/
		if (round = res_cur_size % 4)
		  res_cur_size += 4 -round;
		if (sqv->sqltype < 0)
		  sqv->sqltype = -IISQ_CHA_TYPE;
		else
		  sqv->sqltype = IISQ_CHA_TYPE;
		break;
	}  /* end switch  */

        /* now see if the column is invisible; ie, it could be used
        ** as a sort column, but the user does not want to se it.
        */
        invisible[i] = 0;

        strncpy(defName, sqv->sqlname.sqlnamec, sqv->sqlname.sqlnamel);
        defName[sqv->sqlname.sqlnamel] = '\0';
        strcat(defName, "_invisible");

        if ( defValue = (char *)app_getdefine(defName))
        {
          invisible[i] = 1;
        };

	/* figure how to space the titles for each column; use
	** the maximum of the name of the title, and the actual
	** width of the cloumn.
	*/
	
	colWidth[i] = max(sqv->sqlname.sqlnamel, sqv->sqllen);
	if (!invisible[i])
	        printf("%-*.*s ", colWidth[i] + MARGIN,
           		sqv->sqlname.sqlnamel, sqv->sqlname.sqlnamec);

	if (sqv->sqltype < 0 )
	{
		res_cur_size += sizeof(short);
          	if (round = res_cur_size % 4)
                  res_cur_size += 4 -round;
	};

  }  /* each column ...	*/

  printf("</PRE>\n\n");

  /* allocate a result buffer ...	*/
  
  if (res_buf.res_length >0 && res_buf.res_length < res_cur_size)
  {
	free(res_buf.res_data);
	res_buf.res_length = 0;
  }

  if (res_buf.res_length == 0)
  {
	res_buf.res_data = Alloc_Mem(res_cur_size, 
				"result data storage area");
	res_buf.res_length = res_cur_size;
  }

  /* ok, for each column, assign buffer address */

  for ( res_cur_size = 0, i = 0; i < sqlda->sqld; i++)
  {
	sqv = &sqlda->sqlvar[i];
	
	if ((base_type = sqv->sqltype) < 0 )
	 	base_type = - base_type;

	sqv->sqldata = (char *)&res_buf.res_data[res_cur_size];
	res_cur_size += sqv->sqllen;

	if (base_type == IISQ_CHA_TYPE)
 	{
		res_cur_size++;
		if (round = res_cur_size % 4)
		  res_cur_size += 4 - round;
	}

	if (sqv->sqltype < 0)
	{
		sqv->sqlind = (short *)&res_buf.res_data[res_cur_size];
		res_cur_size += sizeof(short);
                if (round = res_cur_size % 4)
                  res_cur_size += 4 -round;
        }
	else
	{
		sqv->sqlind = (short *)0;
	}
  } /* each col. */

}



void Print_Row()
{

  int i;
  IISQLVAR *sqv;
  int base_type;

  for (i = 0; i < sqlda->sqld; i++)
  {
	if (invisible[i])
	  continue;

	sqv = &sqlda->sqlvar[i];

	if (sqv->sqlind && *sqv->sqlind < 0)
	{
	   printf("%-*s ", colWidth[i]+ MARGIN, nullMessage); 
	}
	else
	{
	  if ( (base_type = sqv->sqltype) < 0)
		base_type = - base_type;

	  switch (base_type)
	  {

	    case IISQ_INT_TYPE:
		printf("%-*d ",  colWidth[i]+MARGIN, *(long *)sqv->sqldata);
		break;

	    case IISQ_FLT_TYPE: 
		printf("%-*g ", colWidth[i]+MARGIN, *(double *)sqv->sqldata);
		break;

	    case IISQ_CHA_TYPE:
		/* if we needto do something different for long strings,
		** then do it here.  I don't know what should be done,
		** so I've left it...
		*/
		printf("%-*.*s ", colWidth[i] + MARGIN, 
				sqv->sqllen, (char *)sqv->sqldata);
		break;

	  } /* end case */

  	} /* not null */


  }

}

void Print_Error()
{

  EXEC SQL BEGIN DECLARE SECTION;
	char error_buff[200];
  EXEC SQL END DECLARE SECTION;

  EXEC SQL INQUIRE_SQL (:error_buff = ERRORTEXT);
  printf("\nSQL Error: \n   %s\n", error_buff);


}


char *Alloc_Mem(mem_size, error_string)
int mem_size;
char *error_string;
{

  char *mem;

  mem = calloc(1, mem_size);
  if (mem) 
	return mem;

  printf("SQL Error:  unable to allocate %s (%d bytes). \n",
		error_string, mem_size);
  EXEC SQL WHENEVER SQLERROR continue;
  EXEC SQL ROLLBACK;
  EXEC SQL disconnect;

}
